@@ -203,8 +203,11 @@ class StringQuery extends QueryType<SearchResult> {
 
   nextMatch(state: EditorState, curFrom: number, curTo: number) {
     let cursor = stringCursor(this.spec, state, curTo, state.doc.length).nextOverlapping()
-    if (cursor.done) cursor = stringCursor(this.spec, state, 0, curFrom).nextOverlapping()
-    return cursor.done ? null : cursor.value
+    if (cursor.done) {
+      let end = Math.min(state.doc.length, curFrom + this.spec.unquoted.length)
+      cursor = stringCursor(this.spec, state, 0, end).nextOverlapping()
+    }
+    return cursor.done || cursor.value.from == curFrom && cursor.value.to == curTo ? null : cursor.value
   }
 
   // Searching in reverse is, rather than implementing an inverted search
@@ -221,8 +224,10 @@ class StringQuery extends QueryType<SearchResult> {
   }
 
   prevMatch(state: EditorState, curFrom: number, curTo: number) {
-    return this.prevMatchInRange(state, 0, curFrom) ||
-      this.prevMatchInRange(state, curTo, state.doc.length)
+    let found = this.prevMatchInRange(state, 0, curFrom)
+    if (!found)
+      found = this.prevMatchInRange(state, Math.max(0, curTo - this.spec.unquoted.length), state.doc.length)
+    return found && (found.from != curFrom || found.to != curTo) ? found : null
   }
 
   getReplacement(_result: SearchResult) { return this.spec.unquote(this.spec.replace) }
@@ -293,11 +298,15 @@ class RegExpQuery extends QueryType<RegExpResult> {
   }
 
   getReplacement(result: RegExpResult) {
-    return this.spec.unquote(this.spec.replace).replace(/\$([$&\d+])/g, (m, i) =>
-      i == "$" ? "$"
-      : i == "&" ? result.match[0]
-      : i != "0" && +i < result.match.length ? result.match[i]
-      : m)
+    return this.spec.unquote(this.spec.replace).replace(/\$([$&]|\d+)/g, (m, i) => {
+      if (i == "&") return result.match[0]
+      if (i == "$") return "$"
+      for (let l = i.length; l > 0; l--) {
+        let n = +i.slice(0, l)
+        if (n > 0 && n < result.match.length) return result.match[n] + i.slice(l)
+      }
+      return m
+    })
   }
 
   matchAll(state: EditorState, limit: number) {
@@ -465,8 +474,9 @@ export const selectSelectionMatches: StateCommand = ({state, dispatch}) => {
 export const replaceNext = searchCommand((view, {query}) => {
   let {state} = view, {from, to} = state.selection.main
   if (state.readOnly) return false
-  let next = query.nextMatch(state, from, from)
-  if (!next) return false
+  let match = query.nextMatch(state, from, from)
+  if (!match) return false
+  let next: SearchResult | null = match
   let changes = [], selection: EditorSelection | undefined, replacement: Text | undefined
   let effects: StateEffect<unknown>[] = []
   if (next.from == from && next.to == to) {
@@ -476,14 +486,16 @@ export const replaceNext = searchCommand((view, {query}) => {
     effects.push(EditorView.announce.of(
       state.phrase("replaced match on line $", state.doc.lineAt(from).number) + "."))
   }
+  let changeSet = view.state.changes(changes)
   if (next) {
-    let off = changes.length == 0 || changes[0].from >= next.to ? 0 : next.to - next.from - replacement!.length
-    selection = EditorSelection.single(next.from - off, next.to - off)
+    selection = EditorSelection.single(next.from, next.to).map(changeSet)
     effects.push(announceMatch(view, next))
     effects.push(state.facet(searchConfigFacet).scrollToMatch(selection.main, view))
   }
   view.dispatch({
-    changes, selection, effects,
+    changes: changeSet,
+    selection,
+    effects,
     userEvent: "input.replace"
   })
   return true
